#!/usr/bin/python
# Copyright (c) 2009 Preston Holmes, David Buxton <david@gasmark6.com>
# 
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following
# conditions:
# 
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
# OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
# HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""A tool for creating transcripts by noting changes made by an installer.
"""
from __future__ import with_statement
from contextlib import contextmanager, nested
from optparse import OptionParser, OptionValueError
from cStringIO import StringIO
from subprocess import CalledProcessError
from time import sleep
import errno
import functools
import itertools
import logging
import os
import re
import shutil
import signal
import subprocess
import sys
import tempfile


logging.basicConfig(format='%(message)s')


USAGE = """
Usage: %prog <PKG> [options]
       %prog <PID> [options]
       %prog --builtin-excludes
""".strip()

DESCRIPTION = (
    "Watches a process or installer package and generates"
    " a transcript of associated filesystem changes.\n\nIf none of --diff, --pkg,"
    " or --radmind is given %prog writes a diff format transcript to standard"
    " out (equivalent to '--diff -')."
    " Must be run as root."
)

LISTPID_CMD = 'ps', 'ax', '-o', 'pid'
FSWATCHER_CMD = 'fsewatcher'
DTRACE_CMD = 'dtrace'
LAUNCHCTL_CMD = 'launchctl'
INSTALLER_CMD = 'installer'
FSDIFF_CMD = 'fsdiff'
PACKAGEMAKER_CMD = '/Developer/usr/bin/packagemaker'
SPOTLIGHT_PLIST = '/System/Library/LaunchDaemons/com.apple.metadata.mds.plist'
# Paths that never appear in a transcript
ALWAYS_EXCLUDE = (
    r'/TemporaryItems/',                # InstallerVise
    r'/\.TemporaryItems/',              # InstallerVise
    r'/VISE Temp Items/',               # InstallerVise
    r'/VISE Temp Items$',               # InstallerVise, how I do not love thee
    r'^/private/var/folders/.*/-Tmp-/', # 10.5 and 10.6 use these
    r'^/private/tmp/',
    r'^/private/var/tmp/',
    r'/\.DS_Store$',
    r'/\.git/',                         # Git
    r'/\.git$',
    r'/\.svn/',                         # Subversion
    r'/\.svn$',
    r'/\.cvsignore$',                   # CVS
    r'/\.cvspass$',
    r'/CVS/',
    r'/CVS$',
    r'/.PKInstallSandbox-tmp/',         # Apple installer temp (10.6, maybe 10.5?)
    r'/.PKInstallSandbox-tmp$',         # Apple installer temp
)


# A dtrace script for snooping process IDs as they are created. The format
# is PID and PPID separated by whitespace on each line. N.B. execsnoop on 10.5
# does not trace syscall::posix_spawn
EXECSNOOP_ARGS = (
    DTRACE_CMD,
    '-q',                       # quiet mode
    '-x', 'switchrate=10hz',    # runtime option
    '-n',                       # the dtrace script itself
    r"""
syscall::execve:return, syscall::posix_spawn:return
{
    printf("%6d %6d\n", pid, ppid);
}
""")

# Nice names for different types of file. Shift mode 4 fields (12 bits)
# to the right to use as a key
FILE_SYMBOLS = {
    001: 'p', # S_IFIFO  0010000  /* named pipe (fifo) */
    002: 'c', # S_IFCHR  0020000  /* character special */
    004: 'd', # S_IFDIR  0040000  /* directory */
    006: 'b', # S_IFBLK  0060000  /* block special */
    010: 'f', # S_IFREG  0100000  /* regular */
    012: 'l', # S_IFLNK  0120000  /* symbolic link */
    014: 's', # S_IFSOCK 0140000  /* socket */
    016: 'w', # S_IFWHT  0160000  /* whiteout */
}

    
@contextmanager
def spotlight_disabled():
    """Context manager that turns off spotlight then turns it back on."""
    args = (LAUNCHCTL_CMD, 'unload', SPOTLIGHT_PLIST)
    logging.info('Stopping spotlight')
    os.spawnvp(os.P_WAIT, args[0], args)
    
    try:
        yield
    # Re-load spotlight even if the block raised an exception
    finally:
        args = (LAUNCHCTL_CMD, 'load', SPOTLIGHT_PLIST)
        logging.info('Starting spotlight')
        os.spawnvp(os.P_WAIT, args[0], args)
        

@contextmanager
def read_program(args, stdout=None, stderr=None):
    """Context manager that runs a subprocess for the program args. The output
    is written to the assignment variable.
    
    You should not read / write the assigned variable (out) within the block.
    """
    logging.info('Starting %s', args[0])
    stdout = stdout or subprocess.PIPE
    stderr = stderr or subprocess.PIPE
    out = StringIO()
    prog = subprocess.Popen(args, stdout=stdout, stderr=stderr)
    
    # Check whether the process failed immediately
    if prog.poll():
        raise CalledProcessError(returncode=prog.returncode, cmd=args[0])
        
    try:
        yield out
    # Stop process and write stdout to file
    finally:
        logging.info('Stopping %s', args[0])
        kill_subprocess(prog)
        progout, progerr = prog.communicate()
        
        # Stop everything if the program did not exit cleanly. We get -15 if
        # the program was terminated by sending SIGTERM
        if prog.returncode not in (0, -15):
            raise CalledProcessError(returncode=prog.returncode, cmd=args[0])

        if progout:
            out.write(progout)
        if progerr:
            out.write(progerr)
        # Rewind the program output for whoever uses it next
        out.seek(0)
        
    
def make_option_parser():
    """Returns a new OptionParser instance.
    
    >>> make_option_parser()    # doctest: +ELLIPSIS
    <optparse.OptionParser instance at ...>
    """
    
    p = OptionParser(usage=USAGE, description=DESCRIPTION)
    p.add_option('-c', '--checksum', metavar='CIPHER',
        help="enable Radmind checksums. See 'fsdiff -V' for a list of ciphers.")
    p.add_option('-d', '--diff', metavar='FILE', help='create a diff transcript.')
    p.add_option('-g', '--grep', metavar='PATTERN', action='append', default=[],
        help='exclude paths matching regular expression.')
    p.add_option('-K', metavar='FILE', help='use FILE for Radmind command file.')
    p.add_option('-p', '--pkg', metavar='PKG', help='create an installer package.')
    p.add_option('-r', '--radmind', metavar='FILE', help='create a Radmind transcript.')
    p.add_option('-v', '--verbose', action='count',
        help='increase verbosity. Can be used multiple times.')
    p.add_option('-x', '--exclude', metavar='PATTERN', action='append', default=[],
        help='exclude file names matching pattern.')
    p.add_option('--stage-dir', metavar='DIR',
        help='use DIR as staging directory when building a package.')
    p.add_option('--target', metavar='DIR', default='/', help='installation target')
    p.add_option('--builtin-excludes', action='store_true', default=False,
        help='print built-in patterns for excluding paths and exit.')
    p.add_option('--no-builtin-excludes', action='store_true', default=False,
        help='do not use built-in patterns for excluding paths.')
    return p


def parse_arguments(parser, argv):
    """Returns an options object and a list of position arguments after parsing
    the argv using the OptionParser instance.
    
    >>> p = make_option_parser()
    >>> _, _ = parse_arguments(p, ['example.pkg'])
    >>> parse_arguments(p, ['example.pkg', '--pkg', 'invalid.mpkg'])
    Traceback (most recent call last):
        ...
    OptionValueError: Argument for --pkg must end with '.pkg'
    >>> parse_arguments(p, ['example.pkg', 'another.pkg', '--pkg', 'invalid.mpkg'])
    Traceback (most recent call last):
        ...
    OptionValueError: Too many process IDs or file names to watch.
    >>> parse_arguments(p, ['example.pkg', '--target', '../relative'])
    Traceback (most recent call last):
        ...
    OptionValueError: Argument for --target must be an absolute path.
    """
    opts, args = parser.parse_args(argv)
    
    if opts.builtin_excludes:
        return opts, args
    
    if len(args) > 1:
        raise OptionValueError('Too many process IDs or file names to watch.')
        
    # Check we have a digit or a string. No mind if it is valid or not.
    if len(args) == 0 or not args[0]:
        raise OptionValueError('Missing required process ID or file name.')
    
    # If none of radmind, pkg or diff then send to stdout as diff
    if not opts.radmind and not opts.pkg and not opts.diff:
        opts.diff = '-'
    
    if opts.pkg and not opts.pkg.endswith('.pkg'):
        raise OptionValueError("Argument for --pkg must end with '.pkg'")
    
    # You cannot specify a relative target
    if not os.path.isabs(opts.target):
        raise OptionValueError('Argument for --target must be an absolute path.')
    
    # Fix a target so it does not end with a trailing /
    opts.target = os.path.abspath(opts.target)
    
    return opts, args


def print_builtin_excludes():
    """Prints the list of built-in exclude patterns to standard out."""
    for pattern in ALWAYS_EXCLUDE:
        sys.stdout.write(pattern + '\n')
        
        
def kill_subprocess(prog, sig=None, terminate=True):
    """Send a signal to a process. If terminate is True wait 10 seconds for it
    to stop. If after 10 seconds it has not stopped send SIGKILL.
    
    By default sends SIGTERM to match the behaviour of kill(1).
    """
    if sig is None:
        sig = signal.SIGTERM
        
    if prog.poll() is not None:
        return
        
    os.kill(prog.pid, sig)
    if terminate:
        for x in range(10):
            sleep(1)
            if prog.poll() is not None:
                return
        else:
            logging.warn('Killing process ID %s after waiting 10 seconds', prog.pid)
            os.kill(prog.pid, signal.SIGKILL)


def watch_process(pid):
    """Waits for the process ID pid to exit. Returns the process ID."""
    logging.info('Waiting for process ID %d', pid)
    while True:
        prog = subprocess.Popen(LISTPID_CMD, stdout=subprocess.PIPE)
        out = prog.communicate()[0]
        pid_list = [int(n.strip()) for n in out.split('\n') if n.strip().isdigit()]
        if pid not in pid_list:
            break
        # Longer sleeps mean lower CPU usage and fewer ps events in execsnoop
        sleep(3)
    return pid


def watch_package(pkg, target='/', verbose=False):
    """Installs the pkg using Apple's installer tool. Returns the process ID
    for the installer process.
    """
    args = (INSTALLER_CMD, '-pkg', pkg, '-target', target)
    env = dict(os.environ, CM_BUILD='CM_BUILD', COMMAND_LINE_INSTALL='1')
    logging.info('Installing package %s', pkg)
    if verbose:
        stdout = None
    else:
        stdout = subprocess.PIPE
    prog = subprocess.Popen(args, env=env, stdout=stdout)
    prog.communicate()
    return prog.pid
    

def watch_install_events(waitfor):
    """Returns a triple of installer process ID, a list of process IDs and a
    list of filesystem changes that happened while waiting for waitfor to return.
    """
    # This works around a problem where data was being dropped when using a PIPE
    # and dtrace hanging after being sent SIGTERM which I guess is to do with
    # pipes blocking and creating a race condition.
    fs_events = tempfile.mkstemp()[1]
    pid_events = tempfile.mkstemp()[1]
    
    with nested(open(fs_events, 'wb'), open(pid_events, 'wb')) as (fs_handle, pid_handle):
        with read_program(EXECSNOOP_ARGS, stdout=pid_handle):
            with read_program([FSWATCHER_CMD], stdout=fs_handle):
                # Wait for event logging processes to spin up
                sleep(1)
                logging.warn("Waiting for installation to complete")
                installer_pid = waitfor()
    
    # Return an iterable object
    fs_events = open(fs_events, 'r')
    pid_events = open(pid_events, 'r')
    
    return installer_pid, pid_events, fs_events
    
    
def watch(waitfor):
    """Records the process IDs involved while waiting for function waitfor.
    Returns a transcript of filesystem changes.
    """
    with spotlight_disabled():
        installer_pid, pid_events, fs_events = watch_install_events(waitfor)
        
    # Strip fs changes relating to other processes
    relevant_pids = find_child_pids(pid_events, installer_pid)
    relevant_pids.update([installer_pid])
    fs_events = filter_fsevents_by_pids(fs_events, relevant_pids)
    
    # Return a generator that does diff format
    return reformat_fsevents(fs_events)
    

def filter_fsevents_by_pids(filename, pids):
    """Returns a generator of lines from the output of fswatcher which match
    one of the given process IDs.

    Filename can be a path or a file-like object open for reading.
    
    Each line has 4 fields. The first field is the process ID
    6593	Terminal	FSE_RENAME	/path/to/filename

    # N.B. The test is slightly wrong because it uses spaces in place of tabs
    >>> from StringIO import StringIO
    >>> log = StringIO(
    ... '''110 Finder FSE_CREATE_FILE /Users/david/watchedinstall/test copy.py
    ... 113 Finder FSE_RENAME /Users/david/watchedinstall/test copy.py
    ... 113 Finder FSE_RENAME /Users/david/.Trash/test copy.py
    ... 372 rm FSE_DELETE /Users/david/watchedinstall/test.py
    ... ''')
    >>> pids = set([113, 110])
    >>> for line in filter_fsevents_by_pids(log, pids): print line,
    ... 
    110 Finder FSE_CREATE_FILE /Users/david/watchedinstall/test copy.py
    113 Finder FSE_RENAME /Users/david/watchedinstall/test copy.py
    113 Finder FSE_RENAME /Users/david/.Trash/test copy.py
    """
    if isinstance(filename, basestring):
        filename = open(filename, 'r')
    for line in filename:
        pid = line.split()[0]
        if pid.isdigit() and int(pid) in pids:
            yield line
    
    
def reformat_fsevents(filename):
    """Returns a generator of lines. Each line is a path preceded by one of
    '!+- ' like diff format. The input format is fswatcher's output.
    
    Filename can be a path or any iterable. Note that new-line characters are
    not stripped from the lines.

    Rename events appear as a delete followed by a create.
    
    >>> from StringIO import StringIO
    >>> log = StringIO(
    ... '''113      Finder  FSE_CREATE_FILE /Users/david/watchedinstall/test copy.py
    ... 113 Finder  FSE_RENAME      /Users/david/watchedinstall/test copy.py
    ... 113 Finder  FSE_RENAME      /Users/david/.Trash/test copy.py
    ... 113 Finder  FSE_CONTENT_MODIFIED    /Users/david/watchedinstall/.DS_Store
    ... 372 rm      FSE_DELETE      /Users/david/watchedinstall/test.py
    ... 4703	Xerox Installer	FSE_CREATE_FILE	/Users/david/991776
    ... ''')
    >>> for line in reformat_fsevents(log): print line,
    ... 
    +/Users/david/watchedinstall/test copy.py
    -/Users/david/watchedinstall/test copy.py
    +/Users/david/.Trash/test copy.py
    !/Users/david/watchedinstall/.DS_Store
    -/Users/david/watchedinstall/test.py
    +/Users/david/991776
    """
    actions = dict(
        FSE_RENAME = '+',
        FSE_DELETE = '-',
        FSE_CREATE_FILE = '+',
        FSE_CREATE_DIR = '+',
        FSE_CONTENT_MODIFIED = '!',
        # FSE_STAT_CHANGED, FSE_CHOWN
    )
    # Regex for output of fswatcher
    # 13	syslogd	FSE_CONTENT_MODIFIED	/private/var/log/system.log
    # Your process better not be called FSE_Installer
    EVENT_MATCH = re.compile(r'^(\d+)\s+(.*)\s+(FSE_\S+)\s+(.*\n)$')
    
    if isinstance(filename, basestring):
        filename = open(filename, 'r')

    # Create an iter object so we can consume extra lines on FSE_RENAME
    lines = iter(filename)
    for line in lines:
        match = EVENT_MATCH.search(line)
        if match:
            pid, process, event, filename = match.groups()
            if event in actions:
                # If the action is rename then first return a delete then move
                # straight to the next line
                if event == 'FSE_RENAME':
                    yield actions['FSE_DELETE'] + filename
                    nextline = lines.next()
                    match = EVENT_MATCH.search(nextline)
                    if match:
                        pid, process, event, filename = match.groups()
                yield actions[event] + filename


def child_pids(ppid, pairs):
    """Returns a set of process IDs that are the descendants of ppid.
    # This should find all the odd pids upto 19
    >>> pairs = [(3, 1), (4, 2), (5, 1), (7, 1), (9, 3), (11, 3), (13, 3),
    ... (6, 2), (8, 2), (10, 2), (12, 10), (14, 10), (16, 14),
    ... (15, 13), (17, 13), (19, 1), ]
    >>> pids = list(child_pids(1, pairs))
    >>> pids.sort()
    >>> pids
    [3, 5, 7, 9, 11, 13, 15, 17, 19]
    """
    # This will loop forever if a child is parent to an ancestor
    # Copy the list, it might be a generator (you stupid idiot)
    pairs = list(pairs)
    children = set(c for c, p in pairs if p == ppid)
    for c in set(children):
        children.update(child_pids(c, pairs))
    return children


def find_child_pids(filename, pid):
    """Returns a set of process IDs that were recorded by execsnoop as being
    owned by the given pid.
    
    Filename can be a path or a file-like object open for reading.
    
    N.B. The original pid will NOT be in the resulting set.
    
    # Convert to list and sort because set order may differ from test to test
    >>> from StringIO import StringIO
    >>> s = StringIO(
    ... '''   4808   4807
    ...    4809   4807
    ...    4810   4807
    ...    4811   4810
    ...    4813   4811
    ...    4814   4811
    ...    4816   4815
    ...    4818   4814
    ...    4817   4814
    ... ''')
    >>> pids = list(find_child_pids(s, 4807))
    >>> pids.sort()
    >>> pids
    [4808, 4809, 4810, 4811, 4813, 4814, 4817, 4818]
    """
    # The output of execsnoop is UID, PID, PPID, ARGS separted by whitespace
    # one per line
    assert isinstance(pid, int)
    pairs = make_pid_pairs(filename)
    return child_pids(pid, pairs)


def make_pid_pairs(filename):
    """Returns a generator of pairs of ints by taking the 1st and 2nd fields for
    each line of the given filename.
    
    Filename can be a path or a file-like object open for reading.
    
    >>> from StringIO import StringIO
    >>> s = StringIO(
    ... '''    4808   4807
    ...     4809   4807
    ...     4810   4807
    ...     4811   4810
    ...     4813   4811
    ...     4814   4811
    ...     4816   4815
    ...     4818   4814
    ...     4817   4814
    ... ''')
    >>> pairs = tuple(make_pid_pairs(s))
    >>> len(pairs)
    9
    >>> pairs   #doctest: +NORMALIZE_WHITESPACE
    ([4808, 4807], [4809, 4807], [4810, 4807], [4811, 4810], [4813, 4811],
     [4814, 4811], [4816, 4815], [4818, 4814], [4817, 4814])
    """
    if isinstance(filename, basestring):
        filename = open(filename, 'r')
    lines = [line for line in filename.read().split('\n') if line.strip()]
    return (map(int, line.split()) for line in lines)


def compile_patterns(globs=None, greps=None, casesensitive=False):
    r"""Compiles each pattern and returns a list of regular expression objects.
    Pass a list of globs for shell-style glob patterns or a list of greps for
    regular expressions.
    
    If casesensitive is False the objects are compiled with ignore case.
    
    >>> import re
    >>> re.IGNORECASE
    2
    >>> for p in compile_patterns(globs=['._*', '.DS_Store']):
    ...     p.pattern, p.flags
    ... 
    ('\\.\\_.*$', 2)
    ('\\.DS\\_Store$', 2)
    >>> for p in compile_patterns(globs=['._*', '.DS_Store'], casesensitive=True):
    ...     p.pattern, p.flags    
    ... 
    ('\\.\\_.*$', 0)
    ('\\.DS\\_Store$', 0)
    >>> for p in compile_patterns(greps=['._*', '.DS_Store']):
    ...     p.pattern, p.flags    
    ... 
    ('._*', 2)
    ('.DS_Store', 2)
    """
    from fnmatch import translate
    
    patterns = []
    flags = re.IGNORECASE
    if casesensitive:
        flags = 0
    
    if globs:
        patterns.extend(re.compile(translate(p), flags) for p in globs)
    
    if greps:
        patterns.extend(re.compile(p, flags) for p in greps)

    return patterns


def make_line_filter(patterns, dropmatches=True):
    """Returns a new function that returns False if the argument matches any
    of the regular expression patterns.
    
    When dropmatches is False, matching lines are included. The default is True,
    meaning that matching lines are excluded (i.e. dropmatches reverses the
    behaviour).
    
    >>> lines = ['+/Example.app', '+/Example.app/Contents/Resources/en.lproj', '+/Example.app/Contents/Info.plist']
    >>> func = make_line_filter(['en.lproj$'])
    >>> [l for l in lines if func(l)]
    ['+/Example.app', '+/Example.app/Contents/Info.plist']
    >>> func = make_line_filter(['Example\.app'])
    >>> [l for l in lines if func(l)]
    []
    >>> func = make_line_filter(['Example\.app'], dropmatches=False)
    >>> [l for l in lines if func(l)]
    ['+/Example.app', '+/Example.app/Contents/Resources/en.lproj', '+/Example.app/Contents/Info.plist']
    """
    dropmatches = bool(dropmatches)
    patterns = [re.compile(p) for p in patterns]
    
    def func(line):
        for p in patterns:
            # Strip leading character (diff format)
            if p.search(line[1:]):
                return not dropmatches
        else:
            return dropmatches
            
    return func
    

def strip_path_target(line, target):
    """Re-writes a line containing a path, stripping the leading target pattern.
    
    >>> strip_path_target('/Volumes/Macintosh HD/tmp/example', '/Volumes/Macintosh HD')
    '/tmp/example'
    """
    if target == '/':
        return line

    if line.startswith(target):
        line = line.replace(target, '', 1)
    return line


def strip_diff_target(line, target):
    """Re-writes a line from a Radmind transcript, stripping the leading
    target pattern from paths.

    >>> strip_diff_target('+/Volumes/Macintosh HD/tmp/example', '/Volumes/Macintosh HD')
    '+/tmp/example'
    """
    prefix, path = line[:1], line[1:]
    return prefix + strip_path_target(path, target)
    

def strip_radmind_target(line, target):
    """Re-writes a line from a Radmind transcript, stripping the leading
    target pattern from paths.

    >>> strip_radmind_target('f /Volumes/Macintosh HD/tmp/index.html     0755   501    20 1060302977    5754 -', '/Volumes/Macintosh HD')
    'f /tmp/index.html     0755   501    20 1060302977    5754 -'
    """
    prefix, path = line[:2], line[2:]
    return prefix + strip_path_target(path, target)


def sort_transcript(transcript):
    """Sort the lines in a transcript by paths.
    
    >>> sort_transcript(['+/Library', '-/Applications', ' /usr', '!/System'])
    ['-/Applications', '+/Library', '!/System', ' /usr']
    """
    transcript = list(transcript)
    return sorted(transcript, key=lambda x: x[1:])
    
    
def camelcase(name):
    """Returns a camel-cased version of the name.
    >>> camelcase('illiterate and ungrammatical')
    'IlliterateAndUngrammatical'
    >>> camelcase('AppleWorks')
    'AppleWorks'
    >>> camelcase('MacHTTP')
    'MacHttp'
    """
    uc = lambda x: x.group(1).upper()
    words = re.split(r'((?<=[a-z])[A-Z]+.*)?', name)    # ish...
    return ''.join(re.sub(r'\s*\b(\w)', uc, word.lower()) for word in words)
    
    
def packagemaker(root, out, id=None, title=None):
    """Create an installer package using the contents of root."""
    title = title or os.path.splitext(os.path.basename(out))[0]
    id = id or ('com.example.' + camelcase(title) + '.pkg')
    args = (
        PACKAGEMAKER_CMD,
        '--root', root,
        '--out', out,
        '--id', id,
        '--title', title,
        '--target', '10.4', # pkg format
    )
    subprocess.Popen(args).wait()
    
    
def copy_filename_to_root(root, filename):
    """Copies the filename into the root directory, taking care to copy links
    as links, directories as directories and files as files.
    
    Ignores other kinds of files.
    """
    try:
        file_info = os.lstat(filename)
    except OSError, err:
        if err.errno == errno.ENOENT:
            logging.info('No such file %s', err.filename)
            return
        raise
        
    symbol = FILE_SYMBOLS[file_info.st_mode >> 12]
    dst = os.path.join(root, filename[1:])  # Strip leading slash on filename
    
    if symbol == 'f':
        copy_file(filename, dst)
    elif symbol == 'l':
        copy_symlink(filename, dst)
    elif symbol == 'd':
        copy_dir(filename, dst)


def copy_dir(src, dst):
    """Copy a directory at src to dst, creating parent directories if needed.
    
    Does not copy the contents of directories. Does nothing if the dst
    directory exists.
    """
    if not os.path.exists(dst):
        copy_dir(os.path.dirname(src), os.path.dirname(dst))
        logging.debug('Copying directory %s to %s', src, dst)
        os.mkdir(dst)
        copy_xattr(src, dst)
        copy_stat(src, dst)


def copy_file(src, dst):
    """Copy a file from src to dst preserving Mac information.
    
    Raises ValueError if the src path does not identify a file.
    """
    copy_dir(os.path.dirname(src), os.path.dirname(dst))
    logging.debug('Copying file %s to %s', src, dst)
    shutil.copyfile(src, dst)
    copy_xattr(src, dst)
    copy_stat(src, dst)


def copy_symlink(src, dst):
    """Copy a symlink from src to dst."""
    copy_dir(os.path.dirname(src), os.path.dirname(dst))
    logging.debug('Copying symlink %s to %s', src, dst)
    link = os.readlink(src)
    os.symlink(link, dst)
    

def copy_stat(src, dst):
    """Copy the values of the stat structure for src to dst.
    
    Does not handle extended attributes.
    """
    info = os.lstat(src)
    os.lchown(dst, info.st_uid, info.st_gid)
    os.chmod(dst, info.st_mode)
    os.utime(dst, (info.st_atime, info.st_mtime))


def copy_xattr(src, dst):
    """Copy the extended attributes for src to dst."""
    import xattr
    
    for key, value in xattr.xattr(src).items():
        xattr.setxattr(dst, key, value)
    
    
def radmind_encode(s):
    r"""Returns a string suitable for Radmind transcripts.
    
    >>> radmind_encode(r'/Library/Application Support')
    '/Library/Application\\bSupport'
    """
    # Translate space, tab, newline, carriage return and backslash
    charmap = {' ': r'\b', '\t': r'\t', '\n': r'\n', '\r': r'\r', '\\': '\\\\'}
    return ''.join(charmap.get(c, c) for c in s)


def create_diff(transcript, filename, target='/'):
    r"""Write the transcript to the filename.
    
    >>> out = StringIO()
    >>> create_diff(['+/Library\n', '-/Applications\n', ' /usr\n', '!/System\n'], out)
    >>> out.getvalue()
    '+/Library\n-/Applications\n /usr\n!/System\n'
    >>> out = StringIO()
    >>> create_diff(['+/V/M/Library\n', '-/V/M/Applications\n'], out, target='/V/M')
    >>> out.getvalue()
    '+/Library\n-/Applications\n'
    """
    logging.info('Writing transcript to diff format')
    if isinstance(filename, basestring):
        filename = open(filename, 'wb')
    
    # Re-write lines for the target
    map_func = functools.partial(strip_diff_target, target=target)
    transcript = itertools.imap(map_func, transcript)
    
    for line in transcript:
        filename.write(line)


def create_radmind(transcript, filename, kommand=None, target='/', checksum=None):
    r"""Write the transcript to the filename in Radmind format.
    
    # Test we don't throw exceptions, ignoring errors in the output format.
    >>> out = StringIO()
    >>> create_radmind(['+/mach_kernel\n'], out)
    >>> out.getvalue().startswith('a /mach_kernel')
    True
    """
    logging.info('Writing transcript to Radmind format')
    if isinstance(filename, basestring):
        filename = open(filename, 'wb')

    remove_kommand = False
    if not kommand:
        kommand = tempfile.mkstemp(suffix='.K')[1]
        remove_kommand = True
    
    pattern = radmind_encode(target)
    
    if checksum:
        args = FSDIFF_CMD, '-K', kommand, '-c', checksum
    else:
        args = FSDIFF_CMD, '-K', kommand

    try:
        for line in transcript:
            if line[0] not in '+!':
                continue
                
            if line[-1] == '\n':
                line = line[:-1]    # Strip newline
            line = line[1:] # Strip leading diff symbol
            if os.path.exists(line):
                diffline = subprocess.Popen(args + (line,), stdout=subprocess.PIPE).communicate()[0]
                # Re-write the path accounting for the target and Radmind's
                # string escapes
                diffline = strip_radmind_target(diffline, pattern)
                filename.write(diffline)
    finally:
        if remove_kommand:
            os.unlink(kommand)
            

def create_package(transcript, filename, root=None, target='/'):
    """Write the transcript to the filename in pkg format."""
    logging.info('Creating package from transcript')
    # Create a root and copy all files into it
    delete_root = False
    if not root:
        root = tempfile.mkdtemp()
        delete_root = True
        logging.debug('Using temp root for building package %s', root)
    
    try:
        for line in transcript:
            if line[0] not in '+!':
                continue
                
            if line[-1] == '\n':
                line = line[:-1]    # Strip newline
            line = line[1:] # Strip leading diff symbol
            copy_filename_to_root(root, line)
        # Compensate for the parent directories that are in the target
        packagemaker(root=root + target, out=filename)
    finally:
        if delete_root:
            logging.debug('Deleting temp root %s', root)
            shutil.rmtree(root)
            
    
def main(argv=None):
    argv = argv or sys.argv
    parser = make_option_parser()
    
    try:
        opts, args = parse_arguments(parser, sys.argv[1:])
    except OptionValueError, err:
        parser.error(err)
    
    logger = logging.getLogger()
    if opts.verbose > 1:
        logger.setLevel(logging.DEBUG)
    if opts.verbose == 1:
        logger.setLevel(logging.INFO)
    
    # Print built-in excludes and exit
    if opts.builtin_excludes:
        print_builtin_excludes()
        sys.exit(0)
        
    # Create the fs transcript
    if args[0].isdigit():
        # Watch a process
        func = functools.partial(watch_process, pid=int(args[0]))
    else:
        # Watch a package
        func = functools.partial(watch_package, pkg=args[0],
            verbose=opts.verbose, target=opts.target)
    # The watch function does the actual pid and fs event logging
    try:
        transcript = watch(waitfor=func)
    except CalledProcessError, err:
        parser.error(err)
    
    # Now we have a transcript. Next we filter paths.
    # Ignore any paths not on the target volume
    patterns = compile_patterns(greps=[r'^' + opts.target])
    filter_func = make_line_filter(patterns, dropmatches=False)
    transcript = itertools.ifilter(filter_func, transcript)
    
    # Apply built-in filename filters. This needs to be fixed to allow filters
    # to work when the target is not '/'.
    if not opts.no_builtin_excludes:
        patterns = compile_patterns(greps=ALWAYS_EXCLUDE)
        default_filter_func = make_line_filter(patterns)
        transcript = itertools.ifilter(default_filter_func, transcript)
    
    # Apply user-specified filename filters
    if opts.exclude or opts.grep:
        patterns = compile_patterns(globs=opts.exclude, greps=opts.grep)
        filter_func = make_line_filter(patterns)
        transcript = itertools.ifilter(filter_func, transcript)
    
    # Consume transcript so we can use it more than once
    transcript = list(transcript)
        
    # Sort the transcript by filename
    transcript = sort_transcript(transcript)
    
    # Now we have a filtered transcript we format it for output.
    # If destination is '-' use stdout
    for outformat in ('radmind', 'diff'):
        if getattr(opts, outformat) == '-':
            setattr(opts, outformat, sys.stdout)
            
    if opts.radmind:
        create_radmind(transcript, opts.radmind, kommand=opts.K,
            target=opts.target, checksum=opts.checksum)
    if opts.diff:
        create_diff(transcript, opts.diff, target=opts.target)
    if opts.pkg:
        create_package(transcript, opts.pkg, root=opts.stage_dir, target=opts.target)


if __name__ == "__main__":
    main()
